# SPDX-FileCopyrightText: 2023-present The Bluemira Developers <https://github.com/Fusion-Power-Plant-Framework/bluemira>
#
# SPDX-License-Identifier: LGPL-2.1-or-later

"""The eqdsk package provides a class for reading and writing eqdsk files.

Module contents:

- [`file.py`][eqdsk.file]: The eqdsk file class.
- `cocos`: The eqdsk coordinate system conversion functions.

"""

from eqdsk.file import EQDSKInterface

__all__ = ["EQDSKInterface"]



# SPDX-FileCopyrightText: 2023-present The Bluemira Developers <https://github.com/Fusion-Power-Plant-Framework/bluemira>
#
# SPDX-License-Identifier: LGPL-2.1-or-later
"""Eqdsk logging"""

import logging

logger = logging.getLogger("EQDSK Logger")

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s [%(levelname)8s] %(message)s",
)


def eqdsk_warn(*args, **kwargs):
    """Warning"""
    logger.warning(*args, **kwargs)


def eqdsk_print(*args, **kwargs):
    """Printing"""
    logger.info(*args, **kwargs)



# SPDX-FileCopyrightText: 2023-present The Bluemira Developers <https://github.com/Fusion-Power-Plant-Framework/bluemira>
#
# SPDX-License-Identifier: LGPL-2.1-or-later

import filecmp
import json
from copy import deepcopy
from pathlib import Path
from unittest import mock

import numpy as np
import pytest

from eqdsk.file import EQDSKInterface
from tests._helpers import compare_dicts, get_private_dir, read_strict_geqdsk


def private_files():
    """Get private files"""
    if (pdir := get_private_dir()) is None:
        return []

    file_path = pdir / "equilibria"

    def _cocos(pth):
        pth = pth.as_posix()
        if "jetto" in pth or "COCOS11" in pth:
            return 11
        return 17

    return [
        *((p, "eqdsk", _cocos(p)) for p in file_path.rglob("*eqdsk*")),
        *((p, "json", _cocos(p)) for p in file_path.rglob("*json")),
    ]


class TestEQDSKInterface:
    """Test the EQDSK interface."""

    @classmethod
    def setup_class(cls):
        cls.data_dir = Path(__file__).parent / "test_data"

    def test_read_default_cocos(self, caplog):
        """Read and return the COCOS for the eqdsk."""
        caplog.set_level("INFO")

        eqd_default = EQDSKInterface.from_file(
            self.data_dir / "jetto.eqdsk_out",
        )

        logs = caplog.records
        assert logs[0].levelname == "WARNING"
        assert "(1, 2, 11, 12)" in logs[0].message
        assert all(lg.levelname == "INFO" for lg in logs[1:])
        assert all(
            int(lg.message.split("COCOS")[-1].strip(".")) == ind
            for lg, ind in zip(logs[1:], [1, 11, 11], strict=True)
        )

        assert eqd_default.cocos.index == EQDSKInterface.DEFAULT_COCOS_INDEX

    def test_read_cocos_specified(self, caplog):
        eqd_as_cc_2 = EQDSKInterface.from_file(
            self.data_dir / "jetto.eqdsk_out",
            volt_seconds_per_radian=True,
            clockwise_phi=True,
            to_cocos_index=2,
        )

        logs = caplog.records
        assert logs[0].levelname == "INFO"
        assert int(logs[0].message.split("COCOS")[-1].strip(".")) == 2
        assert eqd_as_cc_2.cocos.index == 2

    def test_no_cocos_access_raise_ValueError(self):
        with pytest.raises(ValueError, match="not yet been identified"):
            EQDSKInterface.from_file(  # noqa: B018
                self.data_dir / "jetto.eqdsk_out",
                no_cocos=True,
            ).cocos

    @pytest.mark.parametrize(
        ("file", "ftype", "ind"),
        [
            ("jetto.eqdsk_out", "eqdsk", 11),
            ("DN-DEMO_eqref.json", "json", 17),
            ("eqref_OOB.json", "json", 17),
        ],
    )
    def test_read_write_doesnt_change_file(self, file, ftype, ind, tmp_path):
        eqd_default = EQDSKInterface.from_file(self.data_dir / file, ind)

        eqd_default.write(tmp_path / "test", file_format=ftype)

        filecmp.cmp(self.data_dir / file, tmp_path / f"test.{ftype}", shallow=False)

    def test_read_wrong_cocos_raises_ValueError(self):
        with pytest.raises(ValueError, match="No convention found"):
            EQDSKInterface.from_file(self.data_dir / "jetto.eqdsk_out", 17)

    @pytest.mark.parametrize(
        ("setup_keys", "end_keys"),
        [
            ({"from_cocos_index": 17, "to_cocos_index": 11}, {"from_cocos_index": 11}),
            ({"no_cocos": True}, {"no_cocos": True}),
        ],
    )
    def test_read_strict_geqds(self, setup_keys, end_keys, tmp_path):
        file = self.data_dir / "DN-DEMO_eqref.json"
        # Create EQDSK file interface and read data to a dict

        eqdsk = EQDSKInterface.from_file(file, **setup_keys)
        d1 = eqdsk.to_dict()

        # Write data read in from test file into a new EQDSK
        # file, with the suffix "_temp"
        name = Path(file).stem + "_temp"
        fname = Path(tmp_path, f"{name}.eqdsk")
        eqdsk.write(fname, file_format="eqdsk")
        d2 = eqdsk.to_dict()

        # Check eqdsk is readable by Fortran readers.
        # This demands stricter adherence to the G-EQDSK
        # format than eqdsk's main reader.
        read_strict_geqdsk(fname)

        # Write data read in from test file into a new JSON
        # file, with the suffix "_temp"
        jname = fname.with_suffix("").with_suffix(".json")
        eqdsk.write(jname, file_format="json")
        d3 = EQDSKInterface.from_file(jname, **end_keys).to_dict()

        # Compare dictionaries to check data hasn't
        # been changed.
        assert compare_dicts(d1, d2, verbose=True)
        assert compare_dicts(d1, d3, verbose=True)
        assert compare_dicts(d2, d3, verbose=True)

    def test_write_with_wrong_length_qsi_raises_ValueError(self, tmp_path):
        file = self.data_dir / "DN-DEMO_eqref.json"
        # Create EQDSK file interface and read data to a dict

        eqdsk = EQDSKInterface.from_file(file, from_cocos_index=17, to_cocos_index=11)
        eqdsk.qpsi = np.ones(2)

        with pytest.raises(ValueError, match="the length"):
            eqdsk.write(
                Path(tmp_path, f"{Path(file).stem}_temp.eqdsk"), file_format="eqdsk"
            )

    @staticmethod
    @pytest.mark.private()
    @pytest.mark.parametrize(("file", "ftype", "ind"), private_files())
    def test_read_write_doesnt_change_file_private(file, ftype, ind, tmp_path):
        path = tmp_path / "private"
        path.mkdir(exist_ok=True)
        eqd_default = EQDSKInterface.from_file(file, ind)

        eqd_default.write(path / "test", file_format=ftype)

        filecmp.cmp(file, path / f"test.{ftype}", shallow=False)

    def test_derived_field_is_calculated_if_not_given(self):
        data_file = Path(self.data_dir, "DN-DEMO_eqref.json")
        with open(data_file) as f:
            eudemo_sof_data = json.load(f)

        mod_sof_data = deepcopy(eudemo_sof_data)
        for field in ["x", "z", "psinorm"]:
            del mod_sof_data[field]

        with mock.patch(
            "pathlib.Path.open", new=mock.mock_open(read_data=json.dumps(mod_sof_data))
        ):
            eqdsk = EQDSKInterface.from_file("/some/file.json")

        np.testing.assert_allclose(eqdsk.x, eudemo_sof_data["x"])
        np.testing.assert_allclose(eqdsk.z, eudemo_sof_data["z"])
        # The calculation used for psinorm has changed since the
        # eudemo_sof_data was created - so we can't compare to that in
        # this case.
        np.testing.assert_allclose(
            eqdsk.psinorm, np.linspace(0, 1, len(eudemo_sof_data["fpol"]))
        )

    def test_read_matches_values_in_file(self):
        eq = EQDSKInterface.from_file(Path(self.data_dir, "jetto.eqdsk_out"))

        assert eq.nz == 151
        assert eq.nx == 151
        assert eq.xdim == pytest.approx(3.14981545)
        assert eq.ncoil == 0
        assert eq.xc.size == 0
        assert eq.nbdry == 72
        np.testing.assert_allclose(
            eq.xbdry[:3], [0.399993127e01, 0.399150254e01, 0.396906908e01]
        )
        np.testing.assert_allclose(
            eq.zbdry[-3:], [-0.507187454e00, -0.240712636e00, 0.263892047e-01]
        )

    def test_failed_read_eqdsk(self):
        with (
            pytest.raises(OSError, match="Could not read"),
            mock.patch("pathlib.Path.open", new=mock.mock_open(read_data="")),
        ):
            EQDSKInterface.from_file(Path(self.data_dir, "jetto.eqdsk_out"), 11)

        with (
            pytest.raises(OSError, match="Should be at least"),
            mock.patch("pathlib.Path.open", new=mock.mock_open(read_data=" ")),
        ):
            EQDSKInterface.from_file(Path(self.data_dir, "jetto.eqdsk_out"), 11)



# SPDX-FileCopyrightText: 2023-present The Bluemira Developers <https://github.com/Fusion-Power-Plant-Framework/bluemira>
#
# SPDX-License-Identifier: LGPL-2.1-or-later

"""Definitions for the COCOS specification."""

from __future__ import annotations

from copy import deepcopy
from dataclasses import dataclass
from enum import Enum, unique
from typing import TYPE_CHECKING, Any

import numpy as np

from eqdsk.log import eqdsk_warn

if TYPE_CHECKING:
    from eqdsk.file import EQDSKInterface


class ZeroOne(Enum):
    """An enum representing the values 0 and 1 for the 2pi exponent of Bp."""

    ZERO = 0
    ONE = 1

    def __sub__(self, other: Any) -> ZeroOne:
        """Return the difference between the value and the other value.

        - If it is another ZeroOne, return the difference of the values.
        - Raise a `TypeError` otherwise.
        """
        if type(other) is ZeroOne:
            return ZeroOne(self.value - other.value)
        raise TypeError(
            f"Cannot subtract {type(other)} from {type(self)}.",
        )


class Sign(Enum):
    """An enum representing the positive or negative sign of
    a COCOS parameter.
    """

    POSITIVE = 1
    NEGATIVE = -1

    def __mul__(self, other: Any):
        """Return the product of the sign with the other value.

        - If it is another Sign, return the product of the values.
        - If it is a number, return the product of the value and the number.
        """
        if type(other) is Sign:
            return Sign(self.value * other.value)
        return self.value * other


@dataclass(frozen=True)
class COCOSParams:
    """The parameters for a single COCOS definition."""

    index: int
    """The COCOS index"""
    exp_Bp: ZeroOne
    """The exponent for Bp, 0 if the poloidal flux is V.s/2pi, 1 otherwise."""
    sign_Bp: Sign
    """The sign of Bp, depends on the sign of Ip and the gradient of psi."""
    sign_R_phi_Z: Sign
    """The sign of (R, phi, Z), positive if theta and phi have
    opposite directions, negative if the same."""
    sign_rho_theta_phi: Sign
    """The sign of (rho, theta, phi), positive if phi (toroidal)
    is CW from the top."""


@unique
class COCOS(Enum):
    """An enum representing the 16 COCOS definitions."""

    C1 = COCOSParams(
        index=1,
        exp_Bp=ZeroOne.ZERO,
        sign_Bp=Sign.POSITIVE,
        sign_R_phi_Z=Sign.POSITIVE,
        sign_rho_theta_phi=Sign.POSITIVE,
    )
    C11 = COCOSParams(
        index=11,
        exp_Bp=ZeroOne.ONE,
        sign_Bp=Sign.POSITIVE,
        sign_R_phi_Z=Sign.POSITIVE,
        sign_rho_theta_phi=Sign.POSITIVE,
    )
    C2 = COCOSParams(
        index=2,
        exp_Bp=ZeroOne.ZERO,
        sign_Bp=Sign.POSITIVE,
        sign_R_phi_Z=Sign.NEGATIVE,
        sign_rho_theta_phi=Sign.POSITIVE,
    )
    C12 = COCOSParams(
        index=12,
        exp_Bp=ZeroOne.ONE,
        sign_Bp=Sign.POSITIVE,
        sign_R_phi_Z=Sign.NEGATIVE,
        sign_rho_theta_phi=Sign.POSITIVE,
    )
    C3 = COCOSParams(
        index=3,
        exp_Bp=ZeroOne.ZERO,
        sign_Bp=Sign.NEGATIVE,
        sign_R_phi_Z=Sign.POSITIVE,
        sign_rho_theta_phi=Sign.NEGATIVE,
    )
    C13 = COCOSParams(
        index=13,
        exp_Bp=ZeroOne.ONE,
        sign_Bp=Sign.NEGATIVE,
        sign_R_phi_Z=Sign.POSITIVE,
        sign_rho_theta_phi=Sign.NEGATIVE,
    )
    C4 = COCOSParams(
        index=4,
        exp_Bp=ZeroOne.ZERO,
        sign_Bp=Sign.NEGATIVE,
        sign_R_phi_Z=Sign.NEGATIVE,
        sign_rho_theta_phi=Sign.NEGATIVE,
    )
    C14 = COCOSParams(
        index=14,
        exp_Bp=ZeroOne.ONE,
        sign_Bp=Sign.NEGATIVE,
        sign_R_phi_Z=Sign.NEGATIVE,
        sign_rho_theta_phi=Sign.NEGATIVE,
    )
    C5 = COCOSParams(
        index=5,
        exp_Bp=ZeroOne.ZERO,
        sign_Bp=Sign.POSITIVE,
        sign_R_phi_Z=Sign.POSITIVE,
        sign_rho_theta_phi=Sign.NEGATIVE,
    )
    C15 = COCOSParams(
        index=15,
        exp_Bp=ZeroOne.ONE,
        sign_Bp=Sign.POSITIVE,
        sign_R_phi_Z=Sign.POSITIVE,
        sign_rho_theta_phi=Sign.NEGATIVE,
    )
    C6 = COCOSParams(
        index=6,
        exp_Bp=ZeroOne.ZERO,
        sign_Bp=Sign.POSITIVE,
        sign_R_phi_Z=Sign.NEGATIVE,
        sign_rho_theta_phi=Sign.NEGATIVE,
    )
    C16 = COCOSParams(
        index=16,
        exp_Bp=ZeroOne.ONE,
        sign_Bp=Sign.POSITIVE,
        sign_R_phi_Z=Sign.NEGATIVE,
        sign_rho_theta_phi=Sign.NEGATIVE,
    )
    C7 = COCOSParams(
        index=7,
        exp_Bp=ZeroOne.ZERO,
        sign_Bp=Sign.NEGATIVE,
        sign_R_phi_Z=Sign.POSITIVE,
        sign_rho_theta_phi=Sign.POSITIVE,
    )
    C17 = COCOSParams(
        index=17,
        exp_Bp=ZeroOne.ONE,
        sign_Bp=Sign.NEGATIVE,
        sign_R_phi_Z=Sign.POSITIVE,
        sign_rho_theta_phi=Sign.POSITIVE,
    )
    C8 = COCOSParams(
        index=8,
        exp_Bp=ZeroOne.ZERO,
        sign_Bp=Sign.NEGATIVE,
        sign_R_phi_Z=Sign.NEGATIVE,
        sign_rho_theta_phi=Sign.POSITIVE,
    )
    C18 = COCOSParams(
        index=18,
        exp_Bp=ZeroOne.ONE,
        sign_Bp=Sign.NEGATIVE,
        sign_R_phi_Z=Sign.NEGATIVE,
        sign_rho_theta_phi=Sign.POSITIVE,
    )

    def __init__(self, c: COCOSParams):
        """Initialise the COCOS enum."""
        # shortcuts COCOCS instance .value access
        self.index = c.index
        self.exp_Bp = c.exp_Bp
        self.sign_Bp = c.sign_Bp
        self.sign_R_phi_Z = c.sign_R_phi_Z
        self.sign_rho_theta_phi = c.sign_rho_theta_phi

    @classmethod
    def with_index(cls, cocos_index: int) -> COCOS:
        """Return the COCOS of the given index."""
        if not (cocos_index in range(1, 9) or cocos_index in range(11, 19)):
            msg = f"Convention number {cocos_index} is not valid. "
            "Must be between 1 and 8 or 11 and 18."
            raise ValueError(msg)
        return next(x for x in cls if x.index == cocos_index)

    @classmethod
    def matching_convention(
        cls,
        exp_Bp: ZeroOne,
        sign_Bp: Sign,
        sign_R_phi_Z: Sign,
        sign_rho_theta_phi: Sign,
    ) -> COCOS:
        """Return the COCOS matching the given parameters."""

        def _match_cocos(c: COCOS) -> bool:
            return all(
                [
                    c.exp_Bp == exp_Bp,
                    c.sign_Bp == sign_Bp,
                    c.sign_R_phi_Z == sign_R_phi_Z,
                    c.sign_rho_theta_phi == sign_rho_theta_phi,
                ],
            )

        return next(filter(_match_cocos, cls))


@dataclass(frozen=True)
class COCOSTransform:
    """The values for a transformation between in and out COCOS definitions."""

    in_cocos: COCOS
    out_cocos: COCOS

    plasma_current: float
    b_toroidal: float
    coil_current: float
    poloidal_current: float
    psi: float
    pprime: float
    ffprime: float
    q: float


def identify_eqdsk(
    eqdsk: EQDSKInterface,
    *,
    clockwise_phi: bool | None = None,
    volt_seconds_per_radian: bool | None = None,
) -> list[COCOS]:
    """Identify the COCOS for the given
    [EQDSKInterface][eqdsk.file.EQDSKInterface].

    Parameters
    ----------
    eqdsk:
        The eqdsk file to identify the COCOS for.
    clockwise_phi:
        Whether phi is clockwise from the top, by default None
        which means either.
    volt_seconds_per_radian:
        Whether the flux is in volt seconds per radian,
        by default None which means either.

    Returns
    -------
        A list of the identified COCOS definitions.
    """
    if eqdsk.qpsi is None:
        eqdsk_warn("WARNING: qpsi is not defined in the eqdsk file. Setting to 1")
        eqdsk.qpsi = np.array([1])

    cw_phi_l = [True, False] if clockwise_phi is None else [clockwise_phi]
    vs_pr_l = (
        [True, False] if volt_seconds_per_radian is None else [volt_seconds_per_radian]
    )

    definitions = [
        identify_cocos(
            plasma_current=eqdsk.cplasma,
            b_toroidal=eqdsk.bcentre,
            psi_at_boundary=eqdsk.psibdry,
            psi_at_mag_axis=eqdsk.psimag,
            q_psi=eqdsk.qpsi,
            phi_clockwise_from_top=cw_phi,
            volt_seconds_per_radian=vs_pr,
        )
        for cw_phi in cw_phi_l
        for vs_pr in vs_pr_l
    ]

    # return sort asc by index
    definitions.sort(key=lambda x: x.index)
    return definitions


def identify_cocos(
    plasma_current: float,
    b_toroidal: float,
    psi_at_boundary: float,
    psi_at_mag_axis: float,
    q_psi: np.ndarray,
    *,
    phi_clockwise_from_top: bool,
    volt_seconds_per_radian: bool,
) -> COCOS:
    """Identify the COCOS for the given values.

    Parameters
    ----------
    plasma_current:
        The plasma current.
    b_toroidal:
        The toroidal magnetic field.
    psi_at_boundary:
        The psi value at the plasma boundary.
    psi_at_mag_axis:
        The psi value at the magnetic axis.
    q_psi:
        The psi q (safety factor) values.
    phi_clockwise_from_top:
        Whether phi is clockwise from the top.
    volt_seconds_per_radian:
        Whether the flux is in volt seconds per radian.

    Returns
    -------
        The identified COCOS convention.

    Raises
    ------
        ValueError: If the sign of qpsi is not consistent across the flux
            surfaces.
    """
    sign_R_phi_Z = Sign.NEGATIVE if phi_clockwise_from_top else Sign.POSITIVE
    exp_Bp = ZeroOne.ZERO if volt_seconds_per_radian else ZeroOne.ONE

    sign_Ip = Sign(np.sign(plasma_current))
    sign_B0 = Sign(np.sign(b_toroidal))
    sign_psi_inc_towards_boundary = Sign(np.sign(psi_at_boundary - psi_at_mag_axis))

    sign_q = np.sign(q_psi)
    if sign_q.min() != sign_q.max():
        raise ValueError(
            "The sign of qpsi is not consistent across the flux surfaces.",
        )
    sign_q = Sign(sign_q.max())

    sign_Bp = sign_Ip * sign_psi_inc_towards_boundary
    sign_rho_theta_phi = sign_Ip * sign_B0 * sign_q

    return COCOS.matching_convention(
        exp_Bp=exp_Bp,
        sign_Bp=sign_Bp,
        sign_R_phi_Z=sign_R_phi_Z,
        sign_rho_theta_phi=sign_rho_theta_phi,
    )


def transform_cocos(from_cocos_index: int, to_cocos_index: int) -> COCOSTransform:
    """Return the transformation needed to transform from one COCOS
    to another.
    """
    in_cocos = COCOS.with_index(from_cocos_index)
    out_cocos = COCOS.with_index(to_cocos_index)

    eff_bp = (out_cocos.sign_Bp * in_cocos.sign_Bp).value
    eff_R_phi_Z = (out_cocos.sign_R_phi_Z * in_cocos.sign_R_phi_Z).value
    eff_rho_theta_phi = (
        out_cocos.sign_rho_theta_phi * in_cocos.sign_rho_theta_phi
    ).value
    eff_exp_bp = (out_cocos.exp_Bp - in_cocos.exp_Bp).value

    pi_factor = (2 * np.pi) ** eff_exp_bp

    return COCOSTransform(
        in_cocos=in_cocos,
        out_cocos=out_cocos,
        plasma_current=eff_R_phi_Z,
        b_toroidal=eff_R_phi_Z,
        coil_current=eff_R_phi_Z,
        poloidal_current=eff_rho_theta_phi,
        psi=eff_bp * eff_R_phi_Z * (1 / pi_factor),
        pprime=eff_bp * eff_R_phi_Z * pi_factor,
        ffprime=eff_bp * eff_R_phi_Z * pi_factor,
        q=eff_rho_theta_phi,
    )


def convert_eqdsk(eqdsk: EQDSKInterface, to_cocos_index: int) -> EQDSKInterface:
    """Convert an eqdsk file to the given COCOS."""
    in_eqdsk = eqdsk
    out_eqdsk = deepcopy(in_eqdsk)

    in_index = eqdsk.cocos.index
    out_index = to_cocos_index

    if in_index == out_index:
        # returns the copy
        return out_eqdsk

    transform = transform_cocos(
        from_cocos_index=in_index,
        to_cocos_index=out_index,
    )

    out_eqdsk.cplasma = transform.plasma_current * in_eqdsk.cplasma
    out_eqdsk.bcentre = transform.b_toroidal * in_eqdsk.bcentre

    out_eqdsk.Ic = transform.coil_current * in_eqdsk.Ic
    out_eqdsk.fpol = transform.poloidal_current * in_eqdsk.fpol

    out_eqdsk.psi = transform.psi * in_eqdsk.psi
    out_eqdsk.psibdry = transform.psi * in_eqdsk.psibdry
    out_eqdsk.psimag = transform.psi * in_eqdsk.psimag

    out_eqdsk.pprime = transform.pprime * in_eqdsk.pprime
    out_eqdsk.ffprime = transform.ffprime * in_eqdsk.ffprime

    if in_eqdsk.qpsi is not None:
        out_eqdsk.qpsi = transform.q * in_eqdsk.qpsi

    # reidentify the eqdsk
    out_eqdsk.identify(
        clockwise_phi=transform.out_cocos.sign_R_phi_Z == Sign.NEGATIVE,
        volt_seconds_per_radian=transform.out_cocos.exp_Bp == ZeroOne.ZERO,
    )
    if out_eqdsk.cocos.index != to_cocos_index:
        raise RuntimeError(
            f"Failed to convert eqdsk to COCOS {to_cocos_index}, "
            f"eqdsk file identifed as COCOS {out_eqdsk.cocos.index} "
            "post conversion.",
        )

    return out_eqdsk



# SPDX-FileCopyrightText: 2023-present The Bluemira Developers <https://github.com/Fusion-Power-Plant-Framework/bluemira>
#
# SPDX-License-Identifier: LGPL-2.1-or-later

from tests._helpers import get_private_dir


def pytest_addoption(parser):
    """
    Adds a custom command line option to pytest.
    """
    parser.addoption(
        "--private",
        action="store_true",
        dest="private",
        default=False,
        help="run tests that use private data",
    )


def pytest_configure(config):
    """
    Configures pytest.
    """
    if config.option.private and get_private_dir() is None:
        raise ValueError("You cannot run private tests. Data directory not found")

    config.option.markexpr = "" if config.option.private else "not private"



# SPDX-FileCopyrightText: 2023-present The Bluemira Developers <https://github.com/Fusion-Power-Plant-Framework/bluemira>
#
# SPDX-License-Identifier: LGPL-2.1-or-later

import operator
from pathlib import Path
from typing import Any

import fortranformat as ff
import numpy as np

from eqdsk.tools import is_num


def get_private_dir():
    """Get private data directory"""
    if (private_data := Path(__file__).parent / "test_data" / "private").is_dir():
        return private_data
    if (
        private_data := Path(__file__).parent.parent.parent / "bluemira-private-data"
    ).is_dir():
        return private_data
    return None


def read_strict_geqdsk(file_path):
    """
    Reads an input EQDSK file in, assuming strict adherence to the
    GEQDSK format. Used to check bluemira outputs can be read by
    external readers.

    Note: The main bluemira GEQDSK reader is more forgiving to
    format variations than this!

    Parameters
    ----------
    file_path: str
        Full path string of the file

    """
    # Create FortranRecordReader objects with the Fortran format
    # edit descriptors to be used to parse the G-EQDSK input.
    f2000 = ff.FortranRecordReader("a48,3i4")
    f2020 = ff.FortranRecordReader("5e16.9")
    f2022 = ff.FortranRecordReader("2i5")
    fCSTM = ff.FortranRecordReader("i5")

    # Define helper function to read in flattened arrays.
    def read_flat_array(fortran_format, array_size):
        """
        Reads in a flat (1D) numpy array from a G-EQDSK file.

        Parameters
        ----------
        fortran_format: ff.FortranRecordReader
            FortranRecordReader object for Fortran format edit descriptor
            to be used to parse the format of each line of the output.
        array_size: int
            Number of elements in array to be read in.

        Returns
        -------
        array: np.array
            1D Numpy array of length array_size populated by elements from
            the GEQDSK file input.
        """
        # Initialise numpy array and read in first line.
        array = np.zeros((array_size,))
        line = fortran_format.read(file.readline())
        # Define a counter to track which column in a line
        # is currently being saved into the array.
        col = 0
        # Populate array. If the column index moves past the
        # end of a line, it is reset to zero and the next line is read.
        for i in range(array_size):
            if col == len(line):
                line = fortran_format.read(file.readline())
                col = 0
            array[i] = line[col]
            col += 1
        return array

    # Open file.
    with open(file_path) as file:
        # Read in data. Variable names are for readability;
        # strict format is as defined at
        # https://w3.pppl.gov/ntcc/TORAY/G_EQDSK.pdf
        _id, _, nx, nz = f2000.read(file.readline())
        _xdim, _zdim, _xcentre, _xgrid1, _zmid = f2020.read(file.readline())
        _xmag, _zmag, _psimag, _psibdry, _bcentre = f2020.read(file.readline())
        _cplasma, _psimag, _, _xmag, _ = f2020.read(file.readline())
        _zmag, _, _psibdry, _, _ = f2020.read(file.readline())
        _fpol = read_flat_array(f2020, nx)
        _pressure = read_flat_array(f2020, nx)
        _ffprime = read_flat_array(f2020, nx)
        _pprime = read_flat_array(f2020, nx)
        _psi = read_flat_array(f2020, nx * nz)
        _qpsi = read_flat_array(f2020, nx)
        nbdry, nlim = f2022.read(file.readline())
        _xbdry_zbdry = read_flat_array(f2020, 2 * nbdry)
        _xlim_zlim = read_flat_array(f2020, 2 * nlim)

        # Read in coil information, as found in the GEQDSK extension
        (ncoil,) = fCSTM.read(file.readline())
        _coil = read_flat_array(f2020, 5 * ncoil)


def compare_dicts(
    d1: dict[str, Any],
    d2: dict[str, Any],
    *,
    almost_equal: bool = False,
    verbose: bool = True,
    rtol: float = 1e-5,
    atol: float = 1e-8,
) -> bool:
    """
    Compares two dictionaries. Will print information about the differences
    between the two to the console. Dictionaries are compared by length, keys,
    and values per common keys

    Parameters
    ----------
    d1:
        The reference dictionary
    d2:
        The dictionary to be compared with the reference
    almost_equal:
        Whether or not to use np.isclose and np.allclose for numbers and arrays
    verbose:
        Whether or not to print to the console
    rtol:
        The relative tolerance parameter, used if ``almost_equal`` is True
    atol:
        The absolute tolerance parameter, used if ``almost_equal`` is True

    Returns
    -------
    Whether or not the dictionaries are the same
    """
    nkey_diff = len(d1) - len(d2)
    k1 = set(d1.keys())
    k2 = set(d2.keys())
    intersect = k1.intersection(k2)
    new_diff = k1 - k2
    old_diff = k2 - k1
    same, different = [], []

    # Define functions to use for comparison in either the array, dict, or
    # numeric cases.
    def dict_eq(value_1, value_2):
        return compare_dicts(
            value_1,
            value_2,
            almost_equal=almost_equal,
            verbose=verbose,
            rtol=rtol,
            atol=atol,
        )

    def array_almost_eq(val1, val2):
        return np.allclose(val1, val2, rtol, atol)

    def num_almost_eq(val1, val2):
        return np.isclose(val1, val2, rtol, atol)

    def array_is_eq(val1, val2):
        return (np.asarray(val1) == np.asarray(val2)).all()

    if almost_equal:
        array_eq = array_almost_eq
        num_eq = num_almost_eq
    else:
        array_eq = array_is_eq
        num_eq = operator.eq

    # Map the comparison functions to the keys based on the type of value in d1.
    comp_map = {
        key: (
            array_eq
            if isinstance(val, np.ndarray | list)
            else (
                dict_eq
                if isinstance(val, dict)
                else num_eq
                if is_num(val)
                else operator.eq
            )
        )
        for key, val in d1.items()
    }

    # Do the comparison
    for k in intersect:
        v1, v2 = d1[k], d2[k]
        try:
            if comp_map[k](v1, v2):
                same.append(k)
            else:
                different.append(k)
        except ValueError:  # One is an array and the other not
            different.append(k)

    the_same = False
    result = "===========================================================\n"
    if nkey_diff != 0:
        compare = "more" if nkey_diff > 0 else "fewer"
        result += f"d1 has {nkey_diff} {compare} keys than d2" + "\n"
    if new_diff != set():
        result += "d1 has the following keys which d2 does not have:\n"
        new_diff = ["\t" + str(i) for i in new_diff]
        result += "\n".join(new_diff) + "\n"
    if old_diff != set():
        result += "d2 has the following keys which d1 does not have:\n"
        old_diff = ["\t" + str(i) for i in old_diff]
        result += "\n".join(old_diff) + "\n"
    if different:
        result += "the following shared keys have different values:\n"
        different = ["\t" + str(i) for i in different]
        result += "\n".join(different) + "\n"
    if nkey_diff == 0 and new_diff == set() and old_diff == set() and different == []:
        the_same = True
    else:
        result += "==========================================================="
        if verbose:
            print(result)  # noqa: T201
    return the_same



# SPDX-FileCopyrightText: 2023-present The Bluemira Developers <https://github.com/Fusion-Power-Plant-Framework/bluemira>
#
# SPDX-License-Identifier: LGPL-2.1-or-later
"""Eqdsk tools"""

from json import JSONEncoder, dumps
from pathlib import Path

import numpy as np
import numpy.typing as npt

from eqdsk.log import eqdsk_warn


class NumpyJSONEncoder(JSONEncoder):
    """A JSON encoder that can handle numpy arrays."""

    def default(self, obj):
        """Override the JSONEncoder default object handling behaviour
        for np.arrays.
        """
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        return super().default(obj)


def json_writer(
    data: dict,
    file: str | Path | None = None,
    *,
    return_output: bool = False,
    cls=NumpyJSONEncoder,
    **kwargs,
) -> str | None:
    """Write json in the bluemria style.

    Parameters
    ----------
    data:
        dictionary to write to json
    file:
        filename to write to
    return_output:
        return the json as a string
    cls:
        json encoder child class
    kwargs:
        all further kwargs passed to the json writer

    """
    if file is None and not return_output:
        eqdsk_warn("No json action to take")
        return None

    if "indent" not in kwargs:
        kwargs["indent"] = 4

    the_json = dumps(data, cls=cls, **kwargs)

    if file is not None:
        file = Path(file)
        if file.suffix != ".json":
            file = file.with_suffix(".json")

        with open(file, "w") as fh:
            fh.write(the_json)
            fh.write("\n")

    if return_output:
        return the_json
    return None


def is_num(thing) -> bool:
    """Determine whether or not the input is a number.

    Parameters
    ----------
    thing:
        The input which we need to determine is a number or not

    Returns
    -------
    num:
        Whether or not the input is a number
    """
    if thing in {True, False}:
        return False
    try:
        thing = floatify(thing)
    except (ValueError, TypeError):
        return False
    else:
        return not np.isnan(thing)


def floatify(x: npt.ArrayLike) -> float:
    """Converts the np array or float into a float by returning
    the first element or the element itself.

    Notes
    -----
    This function aims to avoid numpy warnings for float(x) for >0 rank scalars
    it emulates the functionality of float conversion

    Raises
    ------
    ValueError
        If array like object has more than 1 element
    TypeError
        If object is None
    """
    if x is None:
        raise TypeError("The argument cannot be None")
    return np.asarray(x, dtype=float).item()



# SPDX-FileCopyrightText: 2023-present The Bluemira Developers <https://github.com/Fusion-Power-Plant-Framework/bluemira>
#
# SPDX-License-Identifier: LGPL-2.1-or-later

"""Read and write EQDSK files."""

from __future__ import annotations

import json
import time
from dataclasses import asdict, dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Any

import fortranformat as ff
import numpy as np

from eqdsk.cocos import COCOS, convert_eqdsk, identify_eqdsk
from eqdsk.log import eqdsk_print, eqdsk_warn
from eqdsk.tools import is_num, json_writer

if TYPE_CHECKING:
    from io import TextIOWrapper

EQDSK_EXTENSIONS = [".eqdsk", ".eqdsk_out", ".geqdsk"]


# repr = False to avoid printing the data in __repr__
@dataclass(repr=False)
class EQDSKInterface:
    """Container for data from an EQDSK file.

    Inspired by an EQDSK reader originally developed by B Dudson:
        https://github.com/bendudson/pyTokamak/blob/master/tokamak/formats/geqdsk.py

    The G-EQDSK file format is described here:
        https://fusion.gat.com/conferences/snowmass/working/mfe/physics/p3/equilibria/g_eqdsk_s.pdf

    Notes
    -----
    G-EQDSK is from the 1980's and EQDSK files should generally only be
    read and not written. New equilibria should really just be saved as
    JSON files.

    Poloidal magnetic flux units not enforced here!

    Plasma current direction is not enforced here!
    """

    DEFAULT_COCOS_INDEX = 11

    bcentre: float
    """Vacuum toroidal Magnetic field at the reference radius [T]."""
    cplasma: float
    """Plasma current [A]."""
    dxc: np.ndarray
    """X half-thicknesses of the coils [m]."""
    dzc: np.ndarray
    """Z half-thicknesses of the coils [m]."""
    ffprime: np.ndarray
    """FF' function on 1-D flux grid [m.T^2/V.s/rad]."""
    fpol: np.ndarray
    """Poloidal current function f = R*B on 1-D flux [T.m]."""
    Ic: np.ndarray
    """Coil currents [A]."""
    name: str
    """Name of the equilibrium EQDSK [dimensionless]."""
    nbdry: int
    """Number of boundary points [dimensionless]."""
    ncoil: int
    """Number of coils [dimensionless]."""
    nlim: int
    """Number of limiters [dimensionless]."""
    nx: int
    """Number of grid points in the radial direction [dimensionless]."""
    nz: int
    """Number of grid points in the vertical direction [dimensionless]."""
    pprime: np.ndarray
    """P' function on 1-D flux grid [N/m^2/V.s/rad]."""
    pressure: np.ndarray
    """Plasma pressure function on 1-D flux grid [N/m^2]."""
    psi: np.ndarray
    """Poloidal magnetic flux on the 2-D grid
    [V.s/rad (COCOS 1-8) or V.s (COCOS 11-18)]."""
    psibdry: float
    """Poloidal flux at the magnetic axis [V.s/rad]."""
    psimag: float
    """Poloidal flux at the magnetic axis [V.s/rad]."""
    xbdry: np.ndarray
    """X coordinates of the plasma boundary [m]."""
    xc: np.ndarray
    """X coordinates of the coils [m]."""
    xcentre: float
    """Radius of the reference toroidal magnetic  [m]."""
    xdim: float
    """Horizontal dimension of the spatial grid [m]."""
    xgrid1: float
    """Minimum radius of the spatial grid [m]."""
    xlim: np.ndarray
    """X coordinates of the limiters [m]."""
    xmag: float
    """Radius of the magnetic axis [m]."""
    zbdry: np.ndarray
    """Z coordinates of the plasma boundary [m]."""
    zc: np.ndarray
    """Z coordinates of the coils [m]."""
    zdim: float
    """Vertical dimension of the spatial grid [m]."""
    zlim: np.ndarray
    """Z coordinates of the limiters [m]."""
    zmag: float
    """Z coordinate of the magnetic axis [m]."""
    zmid: float
    """Z coordinate of the middle of the spatial grid [m]."""
    x: np.ndarray | None = None
    """X 1-D vector [m] (calculated if not given)."""
    z: np.ndarray | None = None
    """Z 1-D vector [m] (calculated if not given)."""
    psinorm: np.ndarray | None = None
    """Normalised psi vector [A] (calculated if not given)."""
    qpsi: np.ndarray | None = None
    """Safety factor values on the 1-D flux grid [dimensionless]."""
    file_name: str | None = None
    """The EQDSK file the data originates from."""

    def __post_init__(self):
        """Calculate derived parameters if they're not given."""
        if self.x is None:
            self.x = _derive_x(self.xgrid1, self.xdim, self.nx)
        if self.z is None:
            self.z = _derive_z(self.zmid, self.zdim, self.nz)
        if self.psinorm is None:
            self.psinorm = _derive_psinorm(self.fpol)
        self._cocos = None

    @classmethod
    def from_file(
        cls,
        file_path: str | Path,
        from_cocos_index: int | None = None,
        to_cocos_index: int | None = DEFAULT_COCOS_INDEX,
        *,
        clockwise_phi: bool | None = None,
        volt_seconds_per_radian: bool | None = None,
        no_cocos: bool = False,
    ) -> EQDSKInterface:
        """Create an EQDSKInterface object from a file.

        Parameters
        ----------
        file_path:
            Path to a file of one of the following formats:
                - JSON
                - eqdsk
                - eqdsk_out
                - geqdsk
        clockwise_phi:
            Whether the EQDSK file's phi is clockwise or not.
        volt_seconds_per_radian:
            Whether the EQDSK file's psi is in volt seconds per radian.
        from_cocos_index:
            The COCOS index of the EQDSK file. Used when the determined
            COCOS is ambiguous. Will raise if given and not one of
            the determined COCOS indices.
        to_cocos_index:
            The COCOS index to convert the EQDSK file to.
        no_cocos:
            Whether to return the EQDSK data without identifying
            and converting to `to_cocos_index` COCOS index.

        Returns
        -------
            An instance of this class containing the EQDSK file's data.
        """
        file_path = Path(file_path)
        file_extension = file_path.suffix
        file_name = file_path.name
        if file_extension.lower() in EQDSK_EXTENSIONS:
            inst = cls(file_name=file_name, **_read_eqdsk(file_path))
        elif file_extension.lower() == ".json":
            inst = cls(file_name=file_name, **_read_json(file_path))
        else:
            raise ValueError(f"Unrecognised file format '{file_extension}'.")

        if no_cocos:
            return inst

        inst.identify(
            as_cocos_index=from_cocos_index,
            clockwise_phi=clockwise_phi,
            volt_seconds_per_radian=volt_seconds_per_radian,
        )
        if to_cocos_index is not None:
            inst = inst.as_cocos(to_cocos_index)

        return inst

    @property
    def cocos(self) -> COCOS:
        """Return the COCOS for this eqdsk."""
        if self._cocos is None:
            raise ValueError(
                "The COCOS for this eqdsk has not yet been identified. "
                "The 'identify' method must be called first "
                "before the COCOS can be returned.",
            )
        return self._cocos

    def identify(
        self,
        as_cocos_index: int | None = None,
        *,
        clockwise_phi: bool | None = None,
        volt_seconds_per_radian: bool | None = None,
    ):
        """Identify the COCOS of this eqdsk and set the COCOS attribute.

        Parameters
        ----------
        clockwise_phi:
            Whether the EQDSK file's phi is clockwise or not.
        volt_seconds_per_radian:
            Whether the EQDSK file's psi is in volt seconds per radian.
        as_cocos_index:
            The COCOS index to convert the EQDSK file to. If given,
            the COCOS will be converted to this COCOS index.

        Raises
        ------
        ValueError:
            If as_cocos_index is given but does not match any
            identified COCOS index.
        ValueError:
            If no COCOS can be identified.

        """
        conventions = identify_eqdsk(
            self,
            clockwise_phi=clockwise_phi,
            volt_seconds_per_radian=volt_seconds_per_radian,
        )

        if as_cocos_index is not None:
            matching_conv = [c for c in conventions if c.index == as_cocos_index]
            if not matching_conv:
                raise ValueError(
                    f"No convention found that matches "
                    f"the given COCOS index {as_cocos_index}, "
                    f"from the possible ({', '.join([str(c.index) for c in conventions])}).",  # noqa: E501
                )
            conventions = matching_conv

        conv = conventions[0]
        if len(conventions) != 1:
            eqdsk_warn(
                f"A single COCOS could not be determined, "
                f"found conventions ({', '.join([str(c.index) for c in conventions])}) "
                f"for the EQDSK file. Choosing COCOS {conv.index}.",
            )
        eqdsk_print(f"EQDSK identified as COCOS {conv.index}.")
        self._cocos = conv

    def as_cocos(self, cocos_index: int) -> EQDSKInterface:
        """Return a copy of this eqdsk converted to the given COCOS."""
        if self.cocos.index == cocos_index:
            return self
        eqdsk_print(f"Converting EQDSK to COCOS {cocos_index}.")
        return convert_eqdsk(self, cocos_index)

    def to_dict(self) -> dict:
        """Return a dictionary of the EQDSK data."""
        d = asdict(self)
        # Remove the file name as this is metadata, not EQDSK data
        del d["file_name"]
        return d

    def write(
        self,
        file_path: str,
        file_format: str = "json",
        json_kwargs: dict | None = None,
    ):
        """Write the EQDSK data to file in the given format.

        Parameters
        ----------
        file_path:
            Path to where the file should be written.
        file_format:
            The format to save the file in. One of 'json', 'eqdsk', or
            'geqdsk'.
        json_kwargs:
            Key word arguments to pass to the ``json.dump`` call. Only
            used if ``format`` is 'json'.
        """
        if file_format == "json":
            json_kwargs = {} if json_kwargs is None else json_kwargs
            json_writer(self.to_dict(), file_path, **json_kwargs)
        elif file_format in {"eqdsk", "geqdsk"}:
            eqdsk_warn(
                "You are in the 21st century. "
                "Are you sure you want to be making an EDQSK in this day and age?"
            )
            _write_eqdsk(file_path, self.to_dict())

    def update(self, eqdsk_data: dict[str, Any]):
        """Update this object's data with values from a dictionary.

        Parameters
        ----------
        eqdsk_data:
            A dict containing the new eqdsk data.

        Raises
        ------
         ValueError:
            If a key in `eqdsk_data` does not correspond to an
            attribute of this class.
        """
        for key, value in eqdsk_data.items():
            if hasattr(self, key):
                setattr(self, key, value)
            else:
                raise ValueError(
                    "Cannot update EQDSKInterface from dict. "
                    f"Unrecognised key '{key}'.",
                )


def _read_json(file_path: Path) -> dict[str, Any]:
    with file_path.open() as file:
        data = json.load(file)

    for k, value in data.items():
        if isinstance(value, list):
            data[k] = np.asarray(value)

    # For backward compatibility where 'psinorm' was sometimes 'pnorm'
    if "pnorm" in data:
        if "psinorm" in data:
            del data["pnorm"]
        else:
            data["psinorm"] = data.pop("pnorm")

    return data


def _read_array(tokens, n, name="Unknown") -> np.ndarray:
    data = np.zeros([n])
    try:
        for i in np.arange(n):
            data[i] = float(next(tokens))
    except StopIteration as e:
        raise ValueError(f"Failed reading array {name} of size {n}") from e
    return data


def _read_2d_array(tokens, n_x, n_y, name="Unknown") -> np.ndarray:
    data = np.zeros([n_y, n_x])
    for i in np.arange(n_y):
        data[i, :] = _read_array(tokens, n_x, f"{name}[{i!s}]]")
    return np.transpose(data)


def _eqdsk_generator(file: TextIOWrapper):
    """Transform a file object into a generator, following G-EQDSK number
    conventions.

    Parameters
    ----------
    file:
        The file to read

    Returns
    -------
        The generator of the file handle being read
    """
    while True:
        line = file.readline()
        if not line:
            break

        # Distinguish negative/positive numbers from negative/positive exponent
        if "E" in line or "e" in line:
            line = line.replace("E-", "*")
            line = line.replace("e-", "*")
            line = line.replace("-", " -")
            line = line.replace("*", "e-")
            line = line.replace("E+", "*")
            line = line.replace("e+", "*")
            line = line.replace("+", " ")
            line = line.replace("*", "e+")
        generator_list = line.split()
        yield from generator_list


def _read_eqdsk(file_path: Path) -> dict:
    with file_path.open("r") as file:
        description = file.readline()
        if not description:
            raise OSError(f"Could not read the file '{file}'.")
        description = description.split()

        ints = [v for v in description if is_num(v)]
        if len(ints) < 3:  # noqa: PLR2004
            raise OSError(
                "Should be at least 3 numbers in the first line "
                f"of the EQDSK {file}.",
            )

        data = {}
        n_x = int(ints[-2])
        n_z = int(ints[-1])
        data["name"] = description[0]
        data["nx"] = n_x
        data["nz"] = n_z

        tokens = _eqdsk_generator(file)
        for name in [
            "xdim",
            "zdim",
            "xcentre",
            "xgrid1",
            "zmid",
            "xmag",
            "zmag",
            "psimag",
            "psibdry",
            "bcentre",
            "cplasma",
            "psimag",
            None,
            "xmag",
            None,
            "zmag",
            None,
            "psibdry",
            None,
            None,
        ]:
            if name is not None:  # Lots of dummies and duplication
                data[name] = float(next(tokens))
            else:
                next(tokens)  # Dummy

        for name in ["fpol", "pressure", "ffprime", "pprime"]:
            data[name] = _read_array(tokens, n_x, name)

        data["psi"] = _read_2d_array(tokens, n_x, n_z, "psi")
        data["qpsi"] = _read_array(tokens, n_x, "qpsi")
        nbdry = int(next(tokens))
        nlim = int(next(tokens))
        data["nbdry"] = nbdry
        data["nlim"] = nlim

        xbdry = np.zeros(nbdry)
        zbdry = np.zeros(nbdry)
        for i in range(nbdry):
            xbdry[i] = float(next(tokens))
            zbdry[i] = float(next(tokens))
        data["xbdry"] = xbdry
        data["zbdry"] = zbdry

        xlim = np.zeros(nlim)
        zlim = np.zeros(nlim)
        for i in range(nlim):
            xlim[i] = float(next(tokens))
            zlim[i] = float(next(tokens))
        data["xlim"] = xlim
        data["zlim"] = zlim

        try:
            ncoil = int(next(tokens))
        except StopIteration:  # No coils in file
            ncoil = 0

        x_c = np.zeros(ncoil)
        z_c = np.zeros(ncoil)
        dxc = np.zeros(ncoil)
        dzc = np.zeros(ncoil)
        i_c = np.zeros(ncoil)
        for i in range(ncoil):
            x_c[i] = float(next(tokens))
            z_c[i] = float(next(tokens))
            dxc[i] = float(next(tokens))
            dzc[i] = float(next(tokens))
            i_c[i] = float(next(tokens))
        data["ncoil"] = ncoil
        data["xc"] = x_c
        data["zc"] = z_c
        data["dxc"] = dxc
        data["dzc"] = dzc
        data["Ic"] = i_c

        # Additional utility data
        data["x"] = _derive_x(data["xgrid1"], data["xdim"], data["nx"])
        data["z"] = _derive_z(data["zmid"], data["zdim"], data["nz"])
        data["psinorm"] = _derive_psinorm(data["fpol"])

    return data


def _derive_x(xgrid1, xdim, nx) -> np.ndarray:
    return np.linspace(xgrid1, xgrid1 + xdim, nx)


def _derive_z(zmid, zdim, nz) -> np.ndarray:
    return np.linspace(zmid - zdim / 2, zmid + zdim / 2, nz)


def _derive_psinorm(fpol) -> np.ndarray:
    return np.linspace(0, 1, len(fpol))


def _write_eqdsk(file_path: str | Path, data: dict):
    """Write data out to a text file in G-EQDSK format.

    Parameters
    ----------
    file:
        The full path string of the file to be created
    data:
        Dictionary of EQDSK data.
    """
    file_path = Path(file_path)
    if file_path.suffix not in EQDSK_EXTENSIONS:
        file_path = file_path.with_suffix(".eqdsk")

    with Path(file_path).open("w") as file:

        def write_header(
            fortran_format: ff.FortranRecordWriter,
            id_string: str,
            var_list: list[str],
        ):
            """Write a G-EQDSK header out to file.

            Parameters
            ----------
            fortran_format:
                FortranRecordWriter object for Fortran format edit
                descriptor to be used for header output.
            id_string:
                String containing name of file to be used as identification
                string. Will be trimmed if length exceeds 39 characters,
                so it will fit within the permitted header length of the
                GEQDSK specification when a timestamp is added.
            var_list:
                List of names of keys in EQDSKInterface.data identifying
                variables to add to the header following the id_string.
                Empty strings will be recorded as 0.
            """
            line = [id_string]
            line += [data[v] if v else 0 for v in var_list]
            file.write(fortran_format.write(line))
            file.write("\n")

        def write_line(fortran_format: ff.FortranRecordWriter, var_list: list[str]):
            """Write a line of variable values out to a G-EQDSK file.

            Parameters
            ----------
            fortran_format:
                FortranRecordWriter object for Fortran format edit
                descriptor to be used for the format of the line output.
            var_list:
                List of names of keys in EQDSKInterface.data identifying
                variables to added to the current line.
                Empty strings will be recorded as 0.
            """
            line = [data[v] if v else 0 for v in var_list]
            file.write(fortran_format.write(line))
            file.write("\n")

        def write_array(fortran_format: ff.FortranRecordWriter, array: np.ndarray):
            """Write a numpy array out to a G-EQDSK file.

            Parameters
            ----------
            fortran_format:
                FortranRecordWriter object for Fortran format edit
                descriptor to be used for the format of the line output.
            array:
                Numpy array of variables to be written to file.
                Array will be flattened in column-major (Fortran)
                order if is more than one-dimensional.
            """
            if array.ndim > 1:
                flat_array = array.flatten(order="F")
                file.write(fortran_format.write(flat_array))
            else:
                file.write(fortran_format.write(array))
            file.write("\n")

        # Create id string for file comprising of timestamp and trimmed filename
        # that fits the 48 character limit of strings in EQDSK headers.
        timestamp = time.strftime("%d%m%Y")
        trimmed_name = data["name"][0 : 48 - len(timestamp) - 1]
        file_id_string = f"{trimmed_name}_{timestamp}"

        # Define dummy data for qpsi if it has not been previously defined.
        qpsi = (
            np.ones(data["nx"]) if data["qpsi"] is None else np.atleast_1d(data["qpsi"])
        )

        if len(qpsi) == 1:
            qpsi = np.full(data["nx"], qpsi)
        elif len(qpsi) != data["nx"]:
            raise ValueError(
                "the length of qpsi should be 1 or the number of x grid points"
            )

        # Create array containing coilset information.
        coil = np.zeros(5 * data["ncoil"])
        for i, value in enumerate(["xc", "zc", "dxc", "dzc", "Ic"]):
            coil[i::5] = data[value]

        # Create FortranRecordWriter objects with the Fortran format
        # edit descriptors to be used in the G-EQDSK output.
        f2000 = ff.FortranRecordWriter("a48,3i4")
        f2020 = ff.FortranRecordWriter("5e16.9")
        f2022 = ff.FortranRecordWriter("2i5")
        fCSTM = ff.FortranRecordWriter("i5")

        # Write header in f2000 (6a8,3i4) format.
        write_header(f2000, file_id_string, ["", "nx", "nz"])
        # Write out lines containing floats in f2020 (5e16.9) format.
        write_line(f2020, ["xdim", "zdim", "xcentre", "xgrid1", "zmid"])
        write_line(f2020, ["xmag", "zmag", "psimag", "psibdry", "bcentre"])
        write_line(f2020, ["cplasma", "psimag", "", "xmag", ""])
        write_line(f2020, ["zmag", "", "psibdry", "", ""])
        # Write out arrays in in f2020 (5e16.9) format.
        write_array(f2020, data["fpol"])
        write_array(f2020, data["pressure"])
        write_array(f2020, data["ffprime"])
        write_array(f2020, data["pprime"])
        write_array(f2020, data["psi"])
        write_array(f2020, qpsi)
        # Write out number of boundary points and limiters f2022 (2i5) format.
        write_line(f2022, ["nbdry", "nlim"])
        # Write out boundary point and limiter data as array of ordered pairs.
        write_array(f2020, np.array([data["xbdry"], data["zbdry"]]))
        write_array(f2020, np.array([data["xlim"], data["zlim"]]))

        # Output of coilset information. This is an extension to the
        # regular eqdsk format.
        write_line(fCSTM, ["ncoil"])
        write_array(f2020, coil)
